"""
Base file for all classes.
"""


class Encoder:
    """
    Base for encoders.
    """

    def __init__(self):
        pass

    def get_spikes(self):
        pass

    def get_spike_time(self):
        pass

    def save(self):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2016 Akshay Raj Gollahalli
"""
Reads the data from folders and json files.
"""

import csv
import logging
import os
import re
import sys

import numpy as np
from sklearn.model_selection import train_test_split

from .errors import LengthMismatchError, NotEnoughDataError, SplitDataError
from .utils import log_it

__all__ = ['ReadCSV']


class ReadCSV(object):
    """Reads the CSV files in a given folder.

    If no folder names is give, this program searches for folder ``Data``; If
    no location is given.

    :param str path:
        Location of the directory containing the data files.
    :param bool log: Default False.
        Verbose mode.
    """

    def __init__(self, path, log=False):

        self.yes_no = log
        self.data_folder = os.path.abspath(path) + os.sep

        log_it(self.yes_no)
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.DEBUG)
        self.logger.debug("ReadCSV class called")

        try:
            if os.path.isdir(self.data_folder):
                self.prefixed = [filename for filename in os.listdir(self.data_folder) if
                                 filename.startswith("sam")]
                self.logger.debug("Reading the data")
            else:
                self.logger.debug("%s does not exist", self.data_folder)
        except IOError as err:
            self.logger.exception("Data files not found - %s", err)
            sys.exit(1)

        if len(self.prefixed) == 1:
            raise NotEnoughDataError("There should be more than one sample to continue.")

    def get_samples(self):
        """
        Reads all the sample starting with sam*.csv and returns a Pandas DataFrame.

        :rtype: dict
        :return samples:
            A dictionary of ``samples`` and ``labels`` (this depends if you have labels file).

        >>> data = ReadCSV('Data')
        >>> data.get_samples()
            {
            'labels': [...],
            'samples: array(...)
            }
        """
        self.prefixed.sort(key=natural_keys)  # Sorted with filename and sample number

        abs_path_of_samples = [self.data_folder + self.prefixed for self.prefixed in self.prefixed]
        raw_samples_list = np.asarray(
            [np.genfromtxt(sample, delimiter=',') for sample in abs_path_of_samples])

        labels = self._get_class_labels()

        if labels[0] is not False:
            if len(raw_samples_list) != len(labels):
                LengthMismatchError(
                    "Number of samples ({0}) is not equal to number of labels ({1}).".format(
                        len(raw_samples_list),
                        len(labels)))
            samples = {'samples': raw_samples_list, 'labels': labels}
        else:
            samples = {'samples': raw_samples_list}

        return samples

    def get_split_data(self, split_to=0.5):
        """Split samples into training and testing data.

        The default test size is 0.5 (50%)

        :param float split_to:
            Percentage split training and testing. Defaults to ``0.5``.

        :rtype: dict
        :return:
            Dictionary of ``train_samples``, ``train_labels``, ``test_samples``,
            ``test_labels`` and ``split_percentage``.
        """

        samples = self.get_samples()

        if split_to > 1:
            raise SplitDataError("Split data should be less that 1.0")

        train_samples, train_labels, test_samples, test_labels = train_test_split(
            samples['samples'],
            samples['labels'],
            test_size=split_to)

        self.logger.debug("Data has been split into Training and Testing data")

        return {'train_samples': train_samples, 'train_labels': train_labels,
                'test_samples': test_samples,
                'test_labels': test_labels,
                'split_percentage': split_to}

    def sample_size(self):
        """Returns the length of the sample size.

        :rtype: int
        :return size:
            Length on samples.
        """
        size = len(self.prefixed)
        self.logger.debug("Sample size returned")
        return size

    def time_feature_length(self):
        """
        Returns the time length of a file by counting it's number of columns.

        :rtype: dict
        :return data_dict:
            Dictionary of ``time_length`` and ``feature_length``.
        """
        _file = self.prefixed[0]
        sample = []
        with open(self.data_folder + _file) as csv_file:
            reader = csv.reader(csv_file, delimiter=' ', quotechar='|')
            for col in reader:
                sample.append(col)

        data_dict = {'time_length': len(sample), 'feature_length': len(sample[0][0].split(','))}
        self.logger.debug("Feature length returned")
        return data_dict

    def _get_class_labels(self):
        _labels = []
        if os.path.isfile(self.data_folder + 'tar_class_labels.csv'):
            with open(self.data_folder + 'tar_class_labels.csv', 'r') as csv_file:
                reader = csv.reader(csv_file, delimiter=' ', quotechar='|')
                for row in reader:
                    _labels.append(int(row[0]))
        else:
            _labels.append(False)

        self.logger.debug("Class labels returned")
        return _labels


def atoi(text):
    """Checks if the file names contain numbers.

    :param str text:
        File name.

    :rtype: bool
    :return: True or false based on the digits in text.
    """
    logging.debug("File names checked")
    return int(text) if text.isdigit() else text


def natural_keys(text):
    """
    Splits the number from the file name.

    :param str text:
        File name.
    :rtype: list
    :return:
        Splits the ``text`` if number is present in it.
    """
    logging.debug("File names split")
    return [atoi(num) for num in re.split(r'(\d+)', text)]



import logging
import sys


def log_it(yes=False):
    """
    Global logging, if set to ``True`` the program runs in verbose mode.

    :param bool yes:
         ``True`` will log the outputs.

    """
    if yes:
        root = logging.getLogger()
        root.setLevel(logging.DEBUG)

        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(asctime)s [%(name)s] [%(funcName)s] [%(threadName)-12.12s] "
            "[%(levelname)-5.5s]  %(message)s")
        ch.setFormatter(formatter)
        root.addHandler(ch)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2016 Akshay Raj Gollahalli
"""
Exception used through out the Spikes api.
"""

__all__ = ['SplitDataError', 'NotEnoughDataError',
           'LengthMismatchError']


class SplitDataError(Exception):
    """
    This exception is raised when the data cannot be split into
    it's appropriate training, testing or validation.
    """

    def __init__(self, message, errors=None):
        super(SplitDataError, self).__init__(message)

        self.errors = errors


class NotEnoughDataError(Exception):
    """
    This exception is raised when the required amount of data is not given.
    """

    def __init__(self, message, errors=None):
        super(NotEnoughDataError, self).__init__(message)

        self.errors = errors


class LengthMismatchError(Exception):
    """
    This exception is raised when the length of sample size is not equal to the length of labels.
    """

    def __init__(self, message, errors=None):
        super(LengthMismatchError, self).__init__(message)

        self.errors = errors



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2016 Akshay Raj Gollahalli

"""
Spike Encoders
==============

In this module we will look at some in-built spike encoders.

There are two encoders provided with this library which  can be used as it is or by spiNNaker.

1. Temporal Based Encoding
2. Bens Spiker Algorithm
"""
import operator

import numpy as np
from scipy.signal import firwin

from .base import Encoder

__all__ = ['TBR', 'BSA']


class TBR(Encoder):
    """
    Threshold Based Representation encoding [1]_

    .. [1] Delbruck, T., & Lichtsteiner, P. (2007, May). Fast sensory motor control based on event-based hybrid
       neuromorphic-procedural system. In 2007 IEEE International Symposium on Circuits and Systems (pp. 845-848). IEEE.

    """

    def __init__(self, data, threshold=0.5):
        """
        :param numpy.ndarray data:
            See :py:mod: nucube.utils.ReadCSV
        :param float threshold:
            Threshold to cut off spike detection.
        """

        super(TBR, self).__init__()
        self.data = data
        self.threshold = threshold

    def get_spikes(self):
        """Returns TBR spikes.

        :rtype: numpy.ndarray
        :return: spikes

        >>> spikes = TBR()
        >>> spikes.get_spikes()
            [[[0 0 0 ..., 0 0 0]
            [1 1 1 ..., 1 1 1]
            [0 0 1 ..., 1 0 1]
            ...,
            [0 0 0 ..., 1 0 0]
            [0 0 0 ..., 0 0 1]
            [0 0 0 ..., 0 0 0]]
            |
            [[0 0 0 ..., 0 0 0]
            [0 0 0 ..., 0 0 0]
            [0 1 0 ..., 0 1 1]
            ...,
            [0 0 0 ..., 0 0 0]
            [0 0 0 ..., 0 0 1]
            [1 0 1 ..., 0 0 1]]]

        """

        sample_feature_threshold = self._get_threshold()
        row_diff = self._get_training_element_difference()

        spikes = []
        for idx, sample in enumerate(row_diff):
            spikes.append(np.where(np.greater(sample, sample_feature_threshold[idx]), 1,
                                   np.where(np.less(sample, -sample_feature_threshold[idx]), -1,
                                            0)))

        # np.array([np.sign(sample_feature_threshold[i] - a) for i, a in enumerate(row_diff)])  # Possible answer

        return np.asarray(spikes, dtype=np.int8)

    def get_spike_time(self, offset=100):
        """
        Merges the samples and returns the time at which the spike happens. Spike time always startes with ``1`` and
        increases with :param: offset.

        :param int offset:
            Shifting the spike index to the given number, if the first index is ``0``, then it will always be ``1``
        :rtype: list
        :return: Spike time with an offset.
        """

        reshaped_raw_spikes = np.vstack(self.get_spikes())

        spike_times = []

        for _feature in reshaped_raw_spikes.T:
            _idx_of_ones = np.squeeze(np.add(np.where(_feature == 1), 1))
            if _idx_of_ones[0] == 1:
                spike_times.append(np.insert(np.multiply(_idx_of_ones[1:], offset), 0, 1).tolist())
            else:
                spike_times.append(np.multiply(_idx_of_ones, offset).tolist())

        return spike_times

    def _get_training_element_difference(self):
        """Returns row-by-row difference for each sample.

        :rtype: numpy.ndarray
        :return: row_diff
        """

        train_data = self.data

        row_diff = []
        for index in range(train_data.shape[0]):
            row_diff.append(np.diff(train_data[index], axis=0))

        return np.asarray(row_diff)

    def _get_mean_sd_sample(self):
        """Returns mean and standard deviation of absolute data.

        :rtype: (numpy.ndarray, numpy.ndarray)
        :return: mean, sd
        """

        diff_data = self._get_training_element_difference()
        abs_data = np.abs(diff_data)
        abs_data_shape_range = range(abs_data.shape[0])

        mean = []
        sd = []
        for index in abs_data_shape_range:
            mean.append(np.mean(abs_data[index], axis=0))

        for index in abs_data_shape_range:
            sd.append(np.std(abs_data[index], axis=0))

        return np.asarray(mean), np.asarray(sd)

    def _get_threshold(self):
        """Returns threshold values for each feature of every sample.

        :rtype: ndarray
        :return: threshold
        """

        mean, sd = self._get_mean_sd_sample()

        return np.add(mean, np.multiply(sd, self.threshold))


class BSA(Encoder):
    """
    Bens Spiker Algorithm (BSA) uses Finite Impulse Response (FIR) filter to try a reverse convolution of the
    stimulation (converting spikes to analog values). A principle borrowed from unpublished modified
    Hough Spiker Algorithm (HSA) [2]_.

    .. [2] Schrauwen, B., & Van Campenhout, J. (2003, July). BSA, a fast and accurate spike train encoding scheme.
        In Proceedings of the international joint conference on neural networks (Vol. 4, No. 4, pp. 2825-2830).
        Piscataway, NJ: IEEE.

    """

    def __init__(self, data, threshold=0.5, filter_length=2, cutoff=0.8):
        """

        :param ndarray data:
            See :py:mod: nucube.utils.ReadCSV.get_samples()
        :param float threshold:
            Threshold level for spike. It can also be a list of threshold according to the numbers of input features.
        :param int filter_length:
        :param float cutoff:
        """

        super(BSA, self).__init__()

        self.data = data
        self.threshold = threshold
        self.filter_length = filter_length
        self.cutoff = cutoff

    def get_spikes(self):
        """
        Returns BSA spikes.

        :rtype: ndarray
        :return output:
            Returns an multidimensional ndarray where each ndarray of ndarray is a sample.

        See ReadData on how to setup the files.

        >>> spike_object = BSA()
        >>> spike_object.get_spikes()
        [[[1 1 1 ... 0 0 0]
          [1 1 1 ... 1 1 1]
          [1 1 1 ... 0 1 1]
          ...
          [1 1 1 ... 1 1 1]
          [0 0 0 ... 0 0 0]
          [0 0 0 ... 0 0 0]]
         [[0 1 1 ... 1 1 1]
          [1 1 1 ... 1 1 1]
          [0 1 0 ... 1 0 1]
          ...
          [0 0 0 ... 0 0 0]
          [0 0 0 ... 0 0 0]
          [0 0 0 ... 0 0 0]]]
        """

        sample_size = self.data.shape[0]
        sample_rate = self.data.shape[1]
        features_size = self.data.shape[2]
        fir_filter = self._fir_filter()

        if isinstance(self.threshold, float) or isinstance(self.threshold, int):
            bsa_row_vector = np.ones(
                features_size) * self.threshold  # Transforming to scalar vector
        elif isinstance(self.threshold, list):
            bsa_row_vector = self.threshold
        else:
            raise TypeError("Unknown type given.")

        min_data = np.array([np.min(_sample, axis=0) for _sample in self.data])
        max_data = np.array([np.max(_sample, axis=0) for _sample in self.data])

        min_data_tile = np.array(
            [np.tile(min_data[num], (sample_rate, 1)) for num in range(sample_size)])
        max_data_tile = np.array(
            [np.tile(max_data[num], (sample_rate, 1)) for num in range(sample_size)])

        encoding_data = np.subtract(self.data, min_data_tile)
        encoding_data = np.divide(encoding_data,
                                  list(map(operator.sub, max_data_tile, min_data_tile)))

        output = np.zeros([sample_size, sample_rate, features_size], dtype=np.int8)

        # BSA Algorithm
        for _sample in range(0, sample_size):
            for _feature in range(0, features_size):
                for _idx in range(0, sample_rate - fir_filter['filter_length']):
                    error1 = 0
                    error2 = 0

                    for j in range(0, fir_filter['filter_length']):
                        error1 += abs(
                            encoding_data[_sample][_idx][_feature] - fir_filter['filter_values'][j])
                        error2 += abs(encoding_data[_sample][_idx][_feature])

                    if error1 <= (error2 - bsa_row_vector[_feature]):
                        output[_sample][_idx][_feature] = 1

                        for j in range(0, fir_filter['filter_length']):
                            if _idx + j <= features_size:
                                encoding_data[_sample][_idx][_feature] -= \
                                    fir_filter['filter_values'][j]

        return output

    def get_spike_time(self, offset=100):
        """
        Merges the samples and returns the time at which the spike happens. Spike time always startes with ``1`` and
        increases with :param: offset.

        :param int offset:
            Shifting the spike index to the given number, if the first index is ``0``, then it will always be ``1``
        :rtype: list
        :return: Spike time with an offset.
        """

        reshaped_raw_spikes = np.vstack(self.get_spikes())

        spike_times = []

        for _feature in reshaped_raw_spikes.T:
            _idx_of_ones = np.squeeze(np.add(np.where(_feature == 1), 1))
            if _idx_of_ones[0] == 1:
                spike_times.append(np.insert(np.multiply(_idx_of_ones[1:], offset), 0, 1).tolist())
            else:
                spike_times.append(np.multiply(_idx_of_ones, offset).tolist())

        return spike_times

    def _fir_filter(self):
        """This method returns FIR filter using window method.

        :rtype: dict
        :return flow:
            The dictionary contains the following:

            - ``filter_values``: A list of float.
            - ``filter_length``: The length of ``filter_value``

        """
        filter_values = firwin(self.filter_length, cutoff=self.cutoff)

        flow = {'filter_values': filter_values, 'filter_length': len(filter_values)}

        return flow



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2016 Akshay Raj Gollahalli

import json

version_json = '''
{
 "version": "0.2.9"
}
'''


def get_versions():
    return json.loads(version_json)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2016 Akshay Raj Gollahalli

"""
Reads the data from folders and json files.
"""

from __future__ import print_function

import csv
import logging
import os
import re
import sys

import numpy as np
from sklearn.model_selection import train_test_split

from .utils.errors import LengthMismatchError, NotEnoughDataError, SplitDataError
from .utils.utils import log_it

__all__ = ['ReadCSV']


class ReadCSV(object):
    """Reads the CSV files in a given folder.

    If no folder names is give, this program searches for folder ``Data``; If
    no location is given.

    :param str path: Needed.
        Location of the directory containing the data files.
    :param bool log: Default False.
        Verbose mode.
    """

    def __init__(self, path, log=False):

        self.yes_no = log
        self.data_folder = os.path.abspath(path) + os.sep

        log_it(self.yes_no)
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.DEBUG)
        self.logger.debug("ReadCSV class called")

        try:
            if os.path.isdir(self.data_folder):
                self.prefixed = [filename for filename in os.listdir(self.data_folder) if
                                 filename.startswith("sam")]
                self.logger.debug("Reading the data")
            else:
                self.logger.debug("%s does not exist", self.data_folder)
        except IOError as err:
            self.logger.exception("Data files not found - %s", err)
            sys.exit(1)

        if len(self.prefixed) == 1:
            raise NotEnoughDataError("There should be more than one sample to continue.")

    def get_samples(self):
        """
        Reads all the sample starting with sam*.csv and returns a Pandas DataFrame.

        :rtype: dict
        :return samples:
            A dictionary of ``samples`` and ``labels`` (this depends if you have labels file).

        >>> data = ReadCSV('Data')
        >>> data.get_samples()
            {
            'labels': [...],
            'samples: array(...)
            }
        """
        self.prefixed.sort(key=natural_keys)  # Sorted with filename and sample number

        abs_path_of_samples = [self.data_folder + self.prefixed for self.prefixed in self.prefixed]
        raw_samples_list = np.asarray(
            [np.genfromtxt(sample, delimiter=',') for sample in abs_path_of_samples])

        labels = self._get_class_labels()

        if labels[0] is not False:
            if len(raw_samples_list) != len(labels):
                LengthMismatchError(
                    "Number of samples ({0}) is not equal to number of labels ({1}).".format(
                        len(raw_samples_list),
                        len(labels)))
            samples = {'samples': raw_samples_list, 'labels': labels}
        else:
            samples = {'samples': raw_samples_list}

        return samples

    def get_split_data(self, split_to=0.5):
        """Split samples into training and testing data.

        The default test size is 0.5 (50%)

        :param float split_to:
            Percentage split training and testing. Defaults to ``0.5``.

        :rtype: dict
        :return:
            Dictionary of ``train_samples``, ``train_labels``, ``test_samples``,
            ``test_labels`` and ``split_percentage``.
        """

        samples = self.get_samples()

        if split_to > 1:
            raise SplitDataError("Split data should be less that 1.0")

        train_samples, train_labels, test_samples, test_labels = train_test_split(
            samples['samples'],
            samples['labels'],
            test_size=split_to)

        self.logger.debug("Data has been split into Training and Testing data")

        return {'train_samples': train_samples, 'train_labels': train_labels,
                'test_samples': test_samples,
                'test_labels': test_labels,
                'split_percentage': split_to}

    def sample_size(self):
        """Returns the length of the sample size.

        :rtype: int
        :return size:
            Length on samples.
        """
        size = len(self.prefixed)
        self.logger.debug("Sample size returned")
        return size

    def time_feature_length(self):
        """
        Returns the time length of a file by counting it's number of columns.

        :rtype: dict
        :return data_dict:
            Dictionary of ``time_length`` and ``feature_length``.
        """
        _file = self.prefixed[0]
        sample = []
        with open(self.data_folder + _file) as csv_file:
            reader = csv.reader(csv_file, delimiter=' ', quotechar='|')
            for col in reader:
                sample.append(col)

        data_dict = {'time_length': len(sample), 'feature_length': len(sample[0][0].split(','))}
        self.logger.debug("Feature length returned")
        return data_dict

    def get_feature_names(self):
        """
        Reads feature names from ``feature_names_eeg.txt`` if it is present else creates a
        feature names automatically.

        :rtype: dict
        :return data_dict:
            A dictionary of Python list, which contains ``number_of_features`` and
            ``name_features``.
        """

        names = []
        number_of_features = 0
        if os.path.isfile(self.data_folder + 'feature_names_eeg.txt'):
            try:
                with open(self.data_folder + 'feature_names_eeg.txt', 'r') as _file:
                    data = _file.read()
                names = data.split('\n')
                number_of_features = len(names)
            except IOError as err:
                self.logger.exception("File not found - %s", err)
        else:
            with open(self.data_folder + self.prefixed[0]) as _file:
                number_of_features = len(_file.readline().split(','))
                for num in range(1, number_of_features + 1):
                    names.append("feature {}".format(num))

        return {'number_of_features': number_of_features, 'name_features': names}

    def _get_class_labels(self):
        _labels = []
        if os.path.isfile(self.data_folder + 'tar_class_labels.csv'):
            with open(self.data_folder + 'tar_class_labels.csv', 'r') as csv_file:
                reader = csv.reader(csv_file, delimiter=' ', quotechar='|')
                for row in reader:
                    _labels.append(int(row[0]))
        else:
            _labels.append(False)

        self.logger.debug("Class labels returned")
        return _labels


def atoi(text):
    """Checks if the file names contain numbers.

    :param str text:
        File name.

    :rtype: bool
    :return: True or false based on the digits in text.
    """
    logging.debug("File names checked")
    return int(text) if text.isdigit() else text


def natural_keys(text):
    """
    Splits the number from the file name.

    :param str text:
        File name.
    :rtype: list
    :return:
        Splits the ``text`` if number is present in it.
    """
    logging.debug("File names split")
    return [atoi(num) for num in re.split(r'(\d+)', text)]



import codecs
import os
from setuptools import setup

from spikes.__version__ import get_versions

name = 'pyspikes'
here = os.path.abspath(os.path.dirname(__file__))


def get_long_description(*parts):
    return codecs.open(os.path.join(here, *parts), 'r').read()


def get_requirements(*parts):
    return codecs.open(os.path.join(here, *parts), 'r').read().splitlines()


setup(
    name=name,
    version=get_versions()['version'],
    packages=['spikes', 'spikes.utils'],
    url='https://github.com/akshaybabloo/Spikes',
    license='BSD-3-Clause',
    author='Akshay Raj Gollahalli',
    author_email='akshay@gollahalli.com',
    description='Spiking Neural Network Spike Encoders',
    long_description=get_long_description('README.md'),
    long_description_content_type="text/markdown",
    install_requires=get_requirements('requirements.txt'),
    keywords="Spiking Neural Network spike encoder",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering",
    ]
)
